from .utils import encode_bson, decode_bson
from pika.exceptions import ProbableAccessDeniedError, ConnectionClosedByBroker
from re import match
import pika
import logging
import requests
import json
import uuid
import copy
import sys

LOGGER = logging.getLogger()  # root logger
LOGGER.setLevel(logging.INFO)
stream_handler = logging.StreamHandler(sys.stdout)
LOGGER.addHandler(stream_handler)


class CreateVhostException(Exception):
    pass


class CreatePermissionException(Exception):
    pass


class RabbitmqManagement(object):
    MANAGEMENT_PORT = 15672
    """
        Rabbitmq management api requests
    """

    def __init__(self, host, username, password):
        self.username = username
        self.password = password
        self.host = host
        self.uri = "http://%s:%s" % (self.host, self.MANAGEMENT_PORT)

    def request_with_credentials(self, url, method="get", data=None):
        req = getattr(requests, method)
        response = req(
            url,
            headers={"content-type": "application/json"},
            data=data,
            auth=(self.username, self.password)
        )
        response.raise_for_status()
        return json.loads(response.content)

    @property
    def messages(self):
        url = "%s/api/queues/" % self.uri
        res = self.request_with_credentials(url)
        return res

    def create_vhost(self, vhost):
        url = "%s/api/vhosts/%s" % (self.uri, vhost)
        try:
            self.request_with_credentials(url, method="put")
        except Exception:
            raise CreateVhostException("Attempt to create temporary vhost=%s failed. url=%r" % (vhost, url))

    def create_permission(self, vhost):
        url = "%s/api/permissions/%s/%s" % (self.uri, vhost, self.username)
        try:
            self.request_with_credentials(url, method="put",
                                          data=json.dumps({"configure": ".*", "write": ".*", "read": ".*"}))
        except Exception:
            raise CreatePermissionException("Attempt to configure premissions on vhost=%s failed. url=%r"
                                            % (vhost, url))

    @classmethod
    def create_vhost_and_permission(cls, vhost, *args, **kwargs):
        klass = cls(*args, **kwargs)
        klass.create_vhost(vhost=vhost)
        klass.create_permission(vhost=vhost)


class Queue(object):
    def __init__(self, uri, vhost, queue_name="", routing_key="", exchange="", exchange_type='fanout'):
        uri = uri.replace('<vhost>', vhost)
        self.uri = uri
        self.exchange = exchange
        self.queue_name = queue_name
        self.routing_key = routing_key
        self.exchange_type = exchange_type
        self.connection, self.channel = self.connect(uri)

    def close(self):
        self.connection.close()
        self.channel.close()

    @classmethod
    def connect(cls, uri):
        """
        :param uri: amqp://guest:guest@rabbitmq.marathon.mesos:5672/vhost?socket_timeout=10
        :return: connection, channel
        """
        LOGGER.info('Connection to %s', uri)
        params = pika.URLParameters(uri)
        try:
            connection = pika.BlockingConnection(params)
        except (ProbableAccessDeniedError, ConnectionClosedByBroker) as e:
            try:
                code, message = e
            except TypeError:
                code, message = e.args
            if match('NOT_ALLOWED - vhost \S+ not found', message):
                RabbitmqManagement.create_vhost_and_permission(params.virtual_host, params.host,
                                                               params.credentials.username, params.credentials.password)
            connection = pika.BlockingConnection(params)
        channel = connection.channel()
        return connection, channel

    def configure(self):
        """ Declare exchange and queue then queue bind to exchange
        :return:
        """
        self.exchange_declare()
        self.queue_declare()
        self.queue_bind()

    def queue_declare(self, durable=True, exclusive=False):
        return self.channel.queue_declare(queue=self.queue_name, durable=durable, exclusive=exclusive)

    def exchange_declare(self):
        return self.channel.exchange_declare(exchange=self.exchange, exchange_type=self.exchange_type)

    def queue_bind(self):
        return self.channel.queue_bind(queue=self.queue_name, exchange=self.exchange)

    def publish(self, body, properties=None):
        prop_dict = copy.deepcopy(properties)
        if properties:
            properties = pika.BasicProperties(**properties)
        try:
            self.channel.basic_publish(exchange=self.exchange, routing_key=self.queue_name, body=body,
                                       properties=properties)
        except Exception as cex:
            LOGGER.info('Exception=%s , Re-configuring the publisher' % cex)
            self.connection, self.channel = self.connect(self.uri)
            self.configure()
            self.publish(body, prop_dict)

    def get_message(self):
        method, properties, body = self.channel.basic_get(self.queue_name, auto_ack=True)
        return method, properties, decode_bson(body)


class BaseConsumer(object):
    def __init__(self, queue):
        self.channel = queue.channel
        self.connection = queue.connection
        self.queue_name = queue.queue_name
        self.callback_func = lambda ch, method, properties, body: self.process(ch, method, properties,
                                                                               decode_bson(body))

    def start_consuming(self, prefetch_count=1):
        self.channel.basic_qos(prefetch_count=prefetch_count)
        self.channel.basic_consume(self.queue_name, self.callback_func)
        self.channel.start_consuming()

    def process(self, ch, method, properties, body):
        """ Callback function
        :param body:
        :return:
        """
        raise NotImplementedError()

    def close(self):
        self.channel.close()
        self.connection.close()

    def start(self):
        """ basic_consume and start_consuming operations
        :return:
        """
        try:
            self.start_consuming()
            LOGGER.info('Consuming started.')
        except Exception as e:
            LOGGER.info(e)
        finally:
            self.close()

    def get_and_publish(self):
        """ Consume one message and close connection
        :return:
        """
        self.channel.basic_qos(prefetch_count=1)
        self.channel.basic_consume(self.queue_name, self.callback_func, auto_ack=True)
        self.connection.process_data_events(time_limit=None)
        self.connection.close()

    def stop(self):
        self.channel.stop_consuming()
        LOGGER.info('Consuming stopped.')


class RpcClient(object):
    def __init__(self, uri, vhost, publish_to):
        """
        Rpc server is match with "options.wait_response" flag in body, publish the response to callback_queue
        """
        self.queue = Queue(uri, vhost)
        self.publish_to = publish_to
        self.callback_queue = self.queue.queue_declare(durable=False, exclusive=True).method.queue
        self.queue.channel.basic_consume(
            queue=self.callback_queue,
            on_message_callback=self.on_response,
            auto_ack=True)

    def on_response(self, ch, method, props, body):
        if self.corr_id == props.correlation_id:
            self.response = body

    def publish_and_wait(self, body, wait_response=True, continue_pipe=True):
        body['options'] = {
            "wait_response": wait_response,
            "continue_pipe": continue_pipe
        }
        self.response = None
        self.corr_id = str(uuid.uuid4())
        self.queue.channel.basic_publish(
            exchange='',
            routing_key=self.publish_to,
            body=encode_bson(body),
            properties=pika.BasicProperties(reply_to=self.callback_queue, correlation_id=self.corr_id)
        )
        while self.response is None:
            self.queue.connection.process_data_events()
        return self.response



# coding=utf-8
from bson.errors import InvalidBSON
from bson.codec_options import CodecOptions
from operator import getitem
from elasticsearch import Elasticsearch
from pymongo import MongoClient
import logging
import json
import bson
import functools

LOGGER = logging.getLogger(__name__)

# Dictionary keys for sentiments.
POSITIVE = 'p'
NEGATIVE = 'n'
NEUTRAL = 'o'

SENTIMENT_VALS = {'p': 100, 'n': 10, 'o': 1}


def sentiment_to_field(sentiment):
    if sentiment == +1:
        return POSITIVE
    elif sentiment == -1:
        return NEGATIVE
    elif sentiment == 0:
        return NEUTRAL
    else:
        raise ValueError('%r is not a valid sentiment value' % sentiment)


def field_to_sentiment(field):
    if field == POSITIVE:
        return +1
    elif field == NEGATIVE:
        return -1
    elif field == NEUTRAL:
        return 0
    else:
        raise ValueError('%r is not a valid sentiment field' % field)


def get_polarity_representation(sentiments):
    # Map post's sentiments into four digit representation.
    # If a post has positive and negative sentiments then it is represented
    # as 1101. If a post contains negative and neutral sentiments then
    # its representation is 1011.
    sentiment = 1000
    for polarity in set(sentiments):
        # sometimes polarity is None. Why ???
        if polarity is None:
            polarity = 0
        sentiment += SENTIMENT_VALS[sentiment_to_field(polarity)]

    return sentiment


def decode_bson(body):
    try:
        options = CodecOptions(tz_aware=True)
        item = bson.BSON(body).decode(options)
        return item
    except InvalidBSON as e:
        LOGGER.error(e)
        LOGGER.warn(json.loads(body))
        LOGGER.warn(body)


def encode_bson(body):
    return bson.BSON.encode(body)


def getter_by_dot_separated(_dict, dot_separated_str):
    """ get dict value with dot notation
    :param _dict:
    :param dot_separated_str: key1.key2.key3
    :return:
    """
    return functools.reduce(getitem, dot_separated_str.split('.'), _dict)


class ElasticConn(object):
    def __init__(self, uri):
        self.uri = uri
        self.client = Elasticsearch(self.uri)

    def __enter__(self):
        return self.client

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.transport.close()


class MongoConn(object):
    def __init__(self, uri, db=None, collection=None):
        self.uri = uri
        self.client = MongoClient(uri)
        self.db = None
        self.collection = None
        if db:
            self.db = self.client[db]
            if collection:
                self.collection = self.db[collection]

    def __enter__(self):
        return self.client

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.close()



import calendar
import datetime
import dateutil.parser
import dateutil.relativedelta
import dateutil.rrule
import pytz
import re

ONE_MONTH_IN_DAYS = 30
ONE_DAY_IN_MILLISECONDS = 86400000
HALF_AN_HOUR_IN_SECONDS = 1800


def current_time():
    return as_utc(datetime.datetime.utcnow())


def as_utc(time):
    return as_time_zone(time, pytz.utc)


def as_time_zone(time, time_zone):
    return time_zone.normalize(time_zone.localize(time))


def to_utc(time):
    return to_time_zone(time, pytz.utc)


def date_to_time(date, time_zone=pytz.UTC):
    time = datetime.datetime.combine(date, datetime.time.min)
    if not time_zone:
        return time
    return as_time_zone(time, time_zone)


def as_time_zone_aware(naive_time):
    if type(naive_time) == datetime.date:
        return date_to_time(naive_time, pytz.UTC)
    elif naive_time.tzinfo is None:
        return as_time_zone(naive_time, pytz.UTC)
    else:
        return naive_time


def to_time_zone(time, time_zone):
    return time.astimezone(time_zone)


def to_timestamp(time):
    return calendar.timegm(to_utc(time).timetuple())


def from_timestamp(timestamp):
    return as_utc(datetime.datetime.utcfromtimestamp(timestamp))


def day_of(time):
    return date_to_time(time.date(), time.tzinfo)


def day_before(day, n=1):
    return date_to_time(day.date() - datetime.timedelta(n), day.tzinfo)


def day_after(day, n=1):
    return date_to_time(day.date() + datetime.timedelta(n), day.tzinfo)


def days_after(since, until):
    day = day_of(since)
    until = day_of(until)
    while day < until:
        yield day
        day = day_after(day)


def days_before(until, since):
    day = day_before(day_of(until))
    since = day_of(since)
    while day >= since:
        yield day
        day = day_before(day)


def end_of_day(date):
    return int(
        round(calendar.timegm(date.timetuple()) * 1000)) + 24 * 60 * 60 * 1000


def render_time(time):
    return time.isoformat()


def render_date(time):
    return time.isoformat().split('T')[0]


def parse_time(string):
    return dateutil.parser.parse(string)


class Period(object):
    PERIOD_PATTERN = re.compile(r'^(?P<value>\d+)(?P<unit>[yqmwdHMS])$')

    def __init__(self, value=1, unit='d'):
        if not isinstance(value, int) or value <= 0:
            raise ValueError('invalid period value %r' % value)

        if unit == 'y':
            self.delta = datetime.timedelta(years=value)
        elif unit == 'q':
            self.delta = dateutil.relativedelta.relativedelta(months=3 * value)
        elif unit == 'm':
            self.delta = dateutil.relativedelta.relativedelta(months=value)
        elif unit == 'w':
            self.delta = datetime.timedelta(weeks=value)
        elif unit == 'd':
            self.delta = datetime.timedelta(days=value)
        elif unit == 'H':
            self.delta = datetime.timedelta(hours=value)
        elif unit == 'M':
            self.delta = datetime.timedelta(minutes=value)
        elif unit == 'S':
            self.delta = datetime.timedelta(seconds=value)
        else:
            raise ValueError('invalid period unit %r' % unit)

        self.value = value
        self.unit = unit

    def round_down(self, time):
        if self.unit == 'y':
            return time - dateutil.relativedelta.relativedelta(
                month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
        elif self.unit == 'q':
            time_zone = time.tzinfo
            naive_time = date_to_time(time.date(), time_zone=None)
            quarters = dateutil.rrule.rrule(
                dateutil.rrule.MONTHLY,
                bymonth=(1, 4, 7, 10),
                bysetpos=-1,
                dtstart=datetime.datetime(naive_time.year, 1, 1),
                count=8)
            return as_time_zone(
                quarters.before(
                    naive_time + dateutil.relativedelta.relativedelta(days=1)),
                time_zone)
        elif self.unit == 'm':
            return time - dateutil.relativedelta.relativedelta(
                day=1, hour=0, minute=0, second=0, microsecond=0)
        elif self.unit == 'w':
            return time - dateutil.relativedelta.relativedelta(
                weekday=dateutil.relativedelta.MO(-1),
                hour=0, minute=0, second=0, microsecond=0)
        elif self.unit == 'd':
            return time - dateutil.relativedelta.relativedelta(
                hour=0, minute=0, second=0, microsecond=0)
        elif self.unit == 'H':
            return time - dateutil.relativedelta.relativedelta(
                minute=0, second=0, microsecond=0)
        elif self.unit == 'M':
            return time - dateutil.relativedelta.relativedelta(
                second=0, microsecond=0)
        elif self.unit == 'S':
            return time - dateutil.relativedelta.relativedelta(microsecond=0)

    def seconds(self):
        return self.delta.total_seconds()

    def intervals(self, start_time, end_time=None):
        if end_time is None:
            end_time = current_time()

        start = self.round_down(start_time)
        end = start + self.delta
        while end <= end_time:
            yield start, end
            start = end
            end += self.delta

    def __repr__(self):
        return 'Period(%r, %r)' % (self.value, self.unit)

    def __str__(self):
        return '%d%s' % (self.value, self.unit)

    def __eq__(self, other):
        return self.value == other.value and self.unit == other.unit

    def __hash__(self):
        return hash((self.value, self.unit))

    @classmethod
    def from_string(cls, string):
        match = re.match(cls.PERIOD_PATTERN, string)
        if match is None:
            raise ValueError('invalid period string %r' % string)

        value = int(match.group('value'))
        unit = match.group('unit')

        return Period(value, unit)
